//------------------------------------------------------------------------
//
// Module:    DummyCCD
// Filename:  DummyCCD.cpp
// Purpose:   Virtual CCD camera to demonstrate and test MaxIm DL/CCD
//            plug-in camera interface.  Implements CDummyCCD class.
// Version:   1.0
//
// Copyright (c) 1998-2003 Diffraction Limited
// Written by Douglas B. George
//
//------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////
// USE OF THIS SOFTWARE
//
// This source code is licensed for use in developing a plug-in DLL for controlling 
// CCD cameras under MaxIm CCD only.  This plug-in source code file shall not be 
// distributed in whole or in part without the express written permission of 
// Diffraction Limited,  Executable DLL files built using this source code 
// may be freely distributed.  

#include <stddef.h>
#include <stdlib.h>
#include <conio.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include <time.h>
#include <string.h>
#include "DummyCCD.h"

// Internal constants
const int WIDTH = 768;									// Physical width of CCD, not including overscan
const int HEIGHT = 512;									// Height of array

const double PixelXSize = 9.0;							// Pixel physical dimensions
const double PixelYSize = 9.0;


//////////////////////////////////////////////////////////////////////
// Non-class DLL entry points
//////////////////////////////////////////////////////////////////////

DLLCALL CMaxImCCDPlugIn * NewPlugIn()
{
	// Create a new CMaxImCCDPlugIn-derived object, and return pointer
	return new CDummyCCD();
}

DLLCALL void DeletePlugIn ( CMaxImCCDPlugIn * PlugIn )
{
	// Delete the previously-created object
	CDummyCCD * Plug = (CDummyCCD *) PlugIn;
	delete Plug;
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CDummyCCD::CDummyCCD()
{
	// Initialization that only needs to happen once
	Buffer = NULL;
	BinningX = 1;
	BinningY = 1;
	StopTime = 0;
	SubframeXStart = 0;
	SubframeYStart = 0;
	SubframeXSize = 0;
	SubframeYSize = 0;
	Temperature = 20.0;
	TSetPoint = 20.0;
	Shutter = false;
	Exposing = false;
	Reading = false;
	LightFrame = false;
	NoiseOn = false;
	GuiderErrors = GE_OFF;
	FilterDelayOn = false;
	OffX = 0.;
	OffY = 0.;
}

CDummyCCD::~CDummyCCD()
{
	// Make sure the camera is shut down
	CloseCamera();
}


//////////////////////////////////////////////////////////////////////
// Member functions
//////////////////////////////////////////////////////////////////////
void CDummyCCD::MakeStar ( unsigned short *Buffer, double CenterX, double CenterY, double Size )
{
	// Create an artificial star for testing guider performance (not needed for a real camera!)
	const float Sigma = 3.0f;
	const int Span = unsigned short ( 10 * Sigma );

	// Convert sigma to 1/sigma squared
	float w = 1.0f / (2.0f * Sigma * Sigma);
	
	CenterX -= SubframeXStart;
	CenterY -= SubframeYStart;
	
	// Loop over image area
	int CY = int ( CenterY );
	int CX = int ( CenterX );
	for (int IY = CY - Span; IY < CY + Span; IY++)
	{
		if (IY < 0 || IY > SubframeYSize) continue;
		double y = IY - CenterY;
		for (int IX = CX - Span ; IX < CX + Span; IX++)
		{
			if (IX < 0 || IX > SubframeXSize) continue;
			double x = IX - CenterX;
			double r = w * ((x*x) + (y*y));
			Buffer[ IX + IY * SubframeXSize ] += unsigned short (Size * exp (-r));
		}
	}
}

//////////////////////////////////////////////////////////////////////
// GetParameters tells MaxIm CCD what information to ask the user for
void CDummyCCD::GetParameters ( 
		DialogContents &Contents,				// Controls the setup dialog box
		PBOOL &HasCamera,						// If true, MaxIm CCD will add this to camera and guider lists
		PBOOL &HasFilterWheel,					// If true, MaxIm CCD will add camera's internal filter wheel to list
		PBOOL &HasGuiderRelays					// If true; MaxIm CCD will use camera's autoguider relays
		)
{
	// Enable filter wheel and guider relays display
	HasCamera = true;
	HasFilterWheel = true;
	HasGuiderRelays = true;

	// Set up title, options
	strcpy ( Contents.CameraName, "Simulator" );
	strcpy ( Contents.FilterName, "Simulator" );
	strcpy ( Contents.Copyright, "CCD simulator Version 3.00\nCopyright (c) 1999-2001 Diffraction Limited\nSupport: cyanogen@cyanogen.com" );
//	ASSERT ( strlen ( Contents.FilterName ) < 15 );
//	ASSERT ( strlen ( Contents.CameraName ) < 15 );
//	ASSERT ( strlen ( Contents.Copyright ) < 255 );

	// Set up dialog box initialization
	Contents.UseFilePath = false;
	Contents.NumParameters = 2;

	// Noise simulation control
	strcpy ( Contents.Parameters[ 0 ].ParameterName, "Noise" );
	Contents.Parameters[ 0 ].NumOptions = 2;

	strcpy ( Contents.Parameters[ 0 ].Option[ 0 ].Display, "Off" );
	Contents.Parameters[ 0 ].Option[ 0 ].Value = false;

	strcpy ( Contents.Parameters[ 0 ].Option[ 1 ].Display, "Simulate" );
	Contents.Parameters[ 0 ].Option[ 1 ].Value = true;

	// Guider simulation control
	strcpy ( Contents.Parameters[ 1 ].ParameterName, "Guide Errors" );
	Contents.Parameters[ 1 ].NumOptions = 4;

	strcpy ( Contents.Parameters[ 1 ].Option[ 0 ].Display, "Off" );
	Contents.Parameters[ 1 ].Option[ 0 ].Value = GE_OFF;

	strcpy ( Contents.Parameters[ 1 ].Option[ 1 ].Display, "Sinusoidal" );
	Contents.Parameters[ 1 ].Option[ 1 ].Value = GE_SINUSOIDAL;

	strcpy ( Contents.Parameters[ 1 ].Option[ 2 ].Display, "Random" );
	Contents.Parameters[ 1 ].Option[ 2 ].Value = GE_RANDOM;

	strcpy ( Contents.Parameters[ 1 ].Option[ 3 ].Display, "Both" );
	Contents.Parameters[ 1 ].Option[ 3 ].Value = GE_BOTH;

	// Set up filter wheel
	Contents.NumFilters = 10;
	Contents.NumFilterParameters = 1;

	// Filter rotation delay simulation control
	strcpy ( Contents.FilterParameters[ 0 ].ParameterName, "Delay" );
	Contents.FilterParameters[ 0 ].NumOptions = 2;

	Contents.FilterParameters[ 0 ].NumOptions = 2;
	strcpy ( Contents.FilterParameters[ 0 ].Option[ 0 ].Display, "On" );
	Contents.FilterParameters[ 0 ].Option[ 0 ].Value = 1;
	strcpy ( Contents.FilterParameters[ 0 ].Option[ 1 ].Display, "Off" );
	Contents.FilterParameters[ 0 ].Option[ 1 ].Value = 0;
}


//////////////////////////////////////////////////////////////////////
// OpenCamera initializes the link to the camera and allocates buffer memory
int CDummyCCD::OpenCamera(
	const char * FilePath,					// File path provided by user (if UseFilePath above is true)
	int Param[ NUM_PARAM ]					// 0-based index for each user parameter input
	)
{
	// If we are re-initializing, make sure old storage deleted
	CloseCamera();

	// Include any initialization that needs to happen each time the camera is started
	// from MaxIm CCD's Setup tab

	// Allocate image buffer used during download
	// Important:  if this function fails, you must deallocate this memory (and any other allocated variables)

	Buffer = new unsigned short[ WIDTH * HEIGHT ];
	if (Buffer == NULL) return RS_OutOfMemory;

	// Check whether we want to add noise to the image
	NoiseOn = Param[ 0 ];
	GuiderErrors = GuiderErrorMode ( Param[ 1 ] );

	BinningX = 1;
	BinningY = 1;
	StopTime = 0;
	SubframeXStart = 0;
	SubframeYStart = 0;
	SubframeXSize = 0;
	SubframeYSize = 0;
	Temperature = 20.0;
	TSetPoint = 20.0;
	Shutter = false;
	Exposing = false;
	Reading = false;
	return 0;
}


//////////////////////////////////////////////////////////////////////
// CloseCamera shuts down the link to the camera and deallocates buffer memory
void CDummyCCD::CloseCamera()
{
	// Shut down link to camera, if any

	// Delete image buffer and reset pointer to null 
	delete [] Buffer;
	Buffer = NULL;

	// Delete any other allocated variables 
	// IMPORTANT -- you must set all pointers to NULL when deleted
}


//////////////////////////////////////////////////////////////////////
// GetArraySize returns the number of pixels in X and Y for the CCD array (unbinned)
void CDummyCCD::GetArraySize ( 
	int &XSize,								// Number of array columns (X)
	int &YSize								// Number of array rows (Y)
	)
{
	XSize = WIDTH;
	YSize = HEIGHT;
}


//////////////////////////////////////////////////////////////////////
// GetImageSize returns the number of pixels in each dimension for the current image, 
// plus offset value for subframe; includes binning
void CDummyCCD::GetImageSize ( 
	int &XStart,							// Column offset for subframe (X) (after binning)
	int &YStart,							// Row offset for subframe (Y) (after binning)
	int &XSize,								// Number of image columns (X)
	int &YSize								// Number of image rows (Y)
	)
{
	XStart = SubframeXStart;
	YStart = SubframeYStart;
	XSize = SubframeXSize;
	YSize = SubframeYSize;
}


//////////////////////////////////////////////////////////////////////
// GetPixelAspect returns the size of the pixels in microns (micrometers)
PBOOL CDummyCCD::GetPixelAspect ( 
	double &PixX,							// Pixel physical X dimension in microns
	double &PixY, 							// Pixel physical Y dimension in microns
	PBOOL IncludeBinning						// If TRUE, multiply above by binning factors
	)
{
	PixX = PixelXSize;
	PixY = PixelYSize;
	if (IncludeBinning)
	{
		PixX *= BinningX;
		PixY *= BinningY;
	}
	return true;
}

//////////////////////////////////////////////////////////////////////
// GetImageBuffer returns a pointer to the image buffer allocated by OpenCamera and filled by TransferImage
unsigned short * CDummyCCD::GetImageBuffer()
{
	return Buffer;
}

//////////////////////////////////////////////////////////////////////
// StartExposures starts a CCD exposure
int CDummyCCD::StartExposure ( 
	unsigned int Exposure,					// Exposure duration in 1/100ths of a second
	int XStart,								// X Offset to start of subframe 
	int YStart, 							// Y Offset to start of subframe 
	int NumX, 								// Number of columns for subframe
	int NumY, 								// Number of rows for subframe
	int BinX, 								// X binning factor (all above numbers are in binned pixel units)
	int BinY, 								// Y binning factor
	PBOOL Light, 							// If true open shutter for exposure, if false close shutter
	PBOOL FastReadout, 						// If true we are in focus mode (optionally use faster readout mode)
	PBOOL HoldShutterOpen 					// If true we are in guider mode (optionally leave shutter open for readout)
	)
{
	if ((XStart + NumX) * BinX > WIDTH) return RS_IllegalInput;
	if ((YStart + NumY) * BinY > HEIGHT) return RS_IllegalInput;

	Shutter = Light;
	LightFrame = Light;
	SubframeXStart = XStart;
	SubframeYStart = YStart;
	SubframeXSize = NumX;
	SubframeYSize = NumY;
	Exposing = true;
	Reading = false;
	Pixel = 0;

	StopTime = clock() + Exposure * CLOCKS_PER_SEC / 100;

	return 0;
}

//////////////////////////////////////////////////////////////////////
// AbortExposure aborts the current imaging operation and returns the camera to idle
// (n.b. If camera hardware needs to be reset, GetCameraState should return CCD_FLUSHING until cleared)
void CDummyCCD::AbortExposure()
{
	StopTime = 0;
}

//////////////////////////////////////////////////////////////////////
// TransferImage downloads the image data from the camera
int CDummyCCD::TransferImage ( 
	PBOOL &TransferDone,						// May return periodically to update GUI, true when done
	int &PercentDone						// If returns periodically during download, percentage complete
	)
{
	int Size = SubframeXSize * SubframeYSize;

	// Simulate readout progress
	// Write some dummy data into the array (consumes time and simulates date)
	for (int I = 0; I < Size / 20; I++)
	{
		if (Pixel >= Size) break;
		if (NoiseOn)
		{
			Buffer[ Pixel++ ] = rand() >> 10;
		}
		else
		{
			Buffer[ Pixel++ ] = 0;
		}
	}

	// Note: recommend exiting periodically, not on every line, to avoid slowing down the readout
	// Returning with percentage is recommended, but optional (some cameras don't like random delays during readout)
	TransferDone = false;
	PercentDone = Pixel * 100 / Size;
	if (PercentDone <= 95) return 0;

	// Make sure last bit of array filled
	while (Pixel < Size)
	{
		if (NoiseOn)
		{
			Buffer[ Pixel++ ] = rand() >> 10;
		}
		else
		{
			Buffer[ Pixel++ ] = 0;
		}
	}
	
	if (LightFrame)
	{
		// Simulate a sinusoidal guider error of +/-three pixels with a time constant of four minutes 
		// (typical worm cycle) plus small random variations of < 1/4 pixel
		time_t Tic = clock() % (CLOCKS_PER_SEC * 240);  // Modulo 240 seconds
		double T = double ( Tic ) / CLOCKS_PER_SEC;
		double Delta = 3. * sin ( 2 * 3.141592654f * T / 240. );
		double AbsOff = 0.;
		
		double RandX = double (rand() - RAND_MAX / 4) / RAND_MAX;
		double RandY = double (rand() - RAND_MAX / 4) / RAND_MAX;
		
		switch (GuiderErrors)
		{
		case GE_OFF :
			break;
		case GE_RANDOM :
			OffX += RandX;
			OffY += RandY;
			break;
		case GE_SINUSOIDAL :
			AbsOff = Delta;
			break;
		case GE_BOTH :
			OffX += RandX;
			OffY += RandY;
			AbsOff = Delta;
			break;
		}
		
		// Make artificial stars
		MakeStar ( Buffer, 256. + OffX + AbsOff, 128. + OffY, 32.f );
		MakeStar ( Buffer, 512. + OffX + AbsOff, 128. + OffY, 64.f );
		MakeStar ( Buffer, 384. + OffX + AbsOff, 256. + OffY, 512.f );
		MakeStar ( Buffer, 256. + OffX + AbsOff, 384. + OffY, 1024.f );
		MakeStar ( Buffer, 512. + OffX + AbsOff, 384. + OffY, 2048.f );
	}

	PercentDone = 100;
	TransferDone = true;
	return 0;
}

//////////////////////////////////////////////////////////////////////
// SetTemperature controls the CCD cooler
PBOOL CDummyCCD::SetTemperature ( 
	PBOOL ControlOn,							// If true, controller is turned on, if false, turned off
	PBOOL GoToAmbient,						// GoToAmbient should start a gradual warm-up to ambient
	double Temp								// Desired setpoint when ControlOn = true and GoToAmbient = false
	)
{
	if (!ControlOn)
	{
		TSetPoint = 20.0;
		Temperature = 20.0;
	}
	else if (GoToAmbient)
	{
		TSetPoint = 20.0;
	}
	else
	{
		TSetPoint = Temp;
	}
	return true;
}

//////////////////////////////////////////////////////////////////////
// GetTemperature returns information about the cooler status
int CDummyCCD::GetTemperature ( 
	CoolerState &State,						// Current temperature controller status
	double &Temp,							// Temperature of CCD chip (FLT_MAX if not available)
	double &TempAmbient,					// Ambient temperature if available (FLT_MAX if not)
	unsigned short &Power					// Power level in percent (0-100), return > 100 if not available
	)
{
	Temp = 0.0f;
	TempAmbient = 0.0f;
	State = COOL_ON;						// You must set State = COOL_NOCONTROL the camera has no cooler control
	Power = 101;
	return 0;
}

//////////////////////////////////////////////////////////////////////
// GetCameraStats returns various information about the camera
void CDummyCCD::GetCameraStats(
	int & XSize,							// Number of rows in physical CCD array (no binning)
	int & YSize, 							// Number of columns in physical CCD array (no binning)
	int & MaxBinFactorX,					// Maximum binning factor allowed on X axis
	int & MaxBinFactorY, 					// Maximum binning factor allowed on Y axis
	PBOOL & DifferentAxes, 					// Return true if X and Y binning factors can be different
	PBOOL & HasShutter, 						// Return true if camera has a shutter
	PBOOL & HasCoolerControl, 				// Return true if camera has a controllable cooler setpoint
	PBOOL & PowerOfTwoBinning, 				// Return true if user can select 1/2/4/8/16 (max) binning only
	PBOOL & HasGuiderRelays					// Return true if camera has built-in autoguider output relays
	)
{
	XSize = WIDTH;
	YSize = HEIGHT;
	MaxBinFactorX = 1;
	MaxBinFactorY = 1;
	DifferentAxes = false;
	HasShutter = true;
	HasCoolerControl = true;
	PowerOfTwoBinning = false;
	HasGuiderRelays = true;
}

//////////////////////////////////////////////////////////////////////
// GetCameraState returns the current camera status
void CDummyCCD::GetCameraState ( 
	CameraState &State,						// Camera state; lower enum values are higher priority
	PBOOL &ShutterOpen						// Return true if shutter is currently open
	)
{
	ShutterOpen = Shutter;
	if (Exposing)
	{
		State = CCD_EXPOSING;

		if (clock() > StopTime)
		{
			Shutter = false;
			ShutterOpen = Shutter;
			Exposing = false;
			Reading = true;
			State = CCD_READING;
			StopTime += CLOCKS_PER_SEC * 2;
			return;
		}
	}
	else if (Reading)
	{
		State = CCD_READING;

		if (clock() > StopTime)
		{
			Reading = false;
			State = CCD_IDLE;
			StopTime = 0;
			return;
		}
	}
	else
	{
		State = CCD_IDLE;
	}
}

//////////////////////////////////////////////////////////////////////
// ActivateRelay turns on an autoguider relay output; only one axis is actuated at a time
int CDummyCCD::ActivateRelay ( 
	int X,									// Activate X axis for T = abs ( X/100 ) seconds, direction based on sign
	int Y									// Activate Y axis for T = abs ( X/100 ) seconds, direction based on sign
	)
{
	// Simulate relay motion, different in X and Y
	OffX += X * 0.02;
	OffY += Y * 0.01;
	return 0;
}

//////////////////////////////////////////////////////////////////////
// IsRelayDone returns true if the autoguider relay is inactive, false if active
PBOOL CDummyCCD::IsRelayDone()
{
	return true;
}

//////////////////////////////////////////////////////////////////////
// OpenFilterWheel opens a filter wheel attached to the camera
int CDummyCCD::OpenFilterWheel ( 
	int Param[ NUM_FPARAM ]					// Parameters specified in 
	)
{
	FilterDelayOn = Param[ 0 ];

	// This filter wheel needs to have the camera initialized in order
	// If the filter wheel is independent, don't do this
	if (Buffer == NULL)
	{
		return RS_FilterNeedsCamera;
	}
	else
	{
		return RS_OK;
	}
}

//////////////////////////////////////////////////////////////////////
// CloseFilterWheel disconnects the filter wheel attached to the camera
void CDummyCCD::CloseFilterWheel()
{

}

//////////////////////////////////////////////////////////////////////
// SetFilterWheel sets the filter wheel position 
int CDummyCCD::SetFilterWheel ( 
	int Position							// 0-based index of filter wheel position
	)
{
	if (FilterDelayOn) 
	{
		// Waste cycles for half a second (ugly...)
		time_t Stop = clock() + CLOCKS_PER_SEC / 2;
		while (clock() < Stop);
	}
	return 0;
}

//////////////////////////////////////////////////////////////////////
// IsFilterWheelMoving returns true if the filter wheel is moving, false if stationary
// Note:  Some cameras only move their filters when a StartExposure command is issued; in that case,
// the GetCameraState command must return CCD_FILTERWHEELMOVING until the wheel has stopped
PBOOL CDummyCCD::IsFilterWheelMoving()
{
	// We'll just return false; MaxIm CCD will very briefly show Moving
	return false;
}
